"""Test permutation.py module."""
import numpy as np
import pytest

import pte_stats


class TestPermutationOnesample:
    """Class for testing permutation_onesample()."""

    @pytest.mark.parametrize("n_perm", [100, 1000])
    @pytest.mark.parametrize("value", [0.0, 1.0])
    @pytest.mark.parametrize("two_tailed", [True, False])
    def test_all_data_is_baseline(self, n_perm, value, two_tailed) -> None:
        """Test case where all values are equal to baseline."""
        data = np.ones(shape=(10,), dtype=float) * value
        z, p = pte_stats.permutation_onesample(
            data_a=data, data_b=value, n_perm=n_perm, two_tailed=two_tailed
        )
        assert p == pytest.approx(1.0)
        assert z == pytest.approx(0.0)

    @pytest.mark.parametrize("two_tailed", [True, False])
    def test_zero_diff(self, two_tailed) -> None:
        """Test case where  values are non significant against baseline."""
        n_perm = 10000
        data = np.ones(shape=(20,), dtype=float)
        data[::2] = data[::2] * -1
        z, p = pte_stats.permutation_onesample(
            data_a=data, data_b=0.0, n_perm=n_perm, two_tailed=two_tailed
        )
        print(f"{p=}")
        print(f"{z=}")
        # assert p == pytest.approx(1.0)
        assert z == pytest.approx(0.0)

    @pytest.mark.parametrize("value", [1.0, -1.0])
    @pytest.mark.parametrize("two_tailed", [True, False])
    def test_baseline_zero(self, value, two_tailed):
        """Test against baseline of 0.0."""
        n_perm = 10000
        data = np.ones(shape=(20,), dtype=float) * value
        z, p = pte_stats.permutation_onesample(
            data_a=data, data_b=0.0, n_perm=n_perm, two_tailed=two_tailed
        )
        if not two_tailed and value == -1.0:
            assert p == pytest.approx(1.0)
            assert z == pytest.approx(value)
            return
        assert p == pytest.approx(1 / (n_perm + 1))
        assert z == pytest.approx(abs(value))


class TestPermutationTwosample:
    """Class for testing permutation_twosample."""

    @pytest.mark.parametrize("value", [1.0, -1.0])
    @pytest.mark.parametrize("two_tailed", [True, False])
    def test_baseline_zero(self, value, two_tailed):
        """Test against baseline of 0.0."""
        n_perm = 10000
        data_a = np.ones(shape=(20,), dtype=float) * value
        data_a[::2] = 0.0
        data_b = np.zeros_like(data_a)
        effect_size, p = pte_stats.permutation_twosample(
            data_a=data_a,
            data_b=data_b,
            n_perm=n_perm,
            two_tailed=two_tailed,
        )
        print(f"{p=}")
        print(f"{effect_size=}")


if __name__ == "__main__":
    for two_tailed in (True, False):
        print(f"{two_tailed=}")
        # TestPermutationOnesample().test_zero_diff(two_tailed=two_tailed)
        TestPermutationTwosample().test_baseline_zero(
            value=1.0, two_tailed=two_tailed
        )



"""An open-source software package for statistics with time series. """

__version__ = "0.2.0"

from .cluster import (
    cluster_analysis_1d,
    cluster_analysis_2d,
    cluster_analysis_from_pvals,
    clusters_from_pvals,
    get_clusters_1d,
)
from .permutation import (
    permutation_onesample,
    permutation_twosample,
    permutation_1d,
    permutation_1d_onesample,
    permutation_1d_twosample,
    permutation_2d,
    permutation_2d_onesample,
    permutation_2d_twosample,
    spearmans_rho_permutation,
)
from .timeseries import (
    timeseries_pvals,
    correct_pvals,
    baseline_correct,
    handle_baseline,
    handle_baseline_bytimes,
)



"""Module for cluster-based statistics."""
import numpy as np
from numba import njit
from skimage import measure

import pte_stats


def cluster_analysis_2d(
    data_a: np.ndarray,
    data_b: np.ndarray | int | float | None = None,
    alpha: float = 0.05,
    n_perm: int = 1000,
    only_max_cluster: bool = False,
    n_jobs: int = 1,
) -> tuple[list, list]:
    """Calculate significant clusters and their corresponding p-values."""
    if data_b is None:
        data_b = 0
    if isinstance(data_b, (int, float)):
        one_sample = True
    else:
        one_sample = False

    two_tailed = True

    # Get 2D clusters
    p_values = pte_stats.permutation_2d(
        data_a=data_a,
        data_b=data_b,
        n_perm=n_perm,
        two_tailed=two_tailed,
    )
    p_values_inv = np.asarray(1 - p_values)

    labels, num_clusters = measure.label(
        p_values <= alpha, return_num=True, connectivity=2
    )  # type: ignore

    null_distr = _null_distribution_2d(
        data_a=data_a,
        data_b=data_b,
        alpha=alpha,
        n_perm=n_perm,
        two_tailed=two_tailed,
        one_sample=one_sample,
        n_jobs=n_jobs,
    )

    p_sum_max = 0  # Is only used if only_max_cluster
    clusters = []
    cluster_pvals = []
    for i in range(num_clusters):
        # Cluster labels start at 1
        index_cluster = np.asarray(labels == i + 1).nonzero()
        p_sum = np.sum(p_values_inv[index_cluster])
        p_val = (n_perm - np.sum(p_sum >= null_distr) + 1) / (n_perm + 1)
        if p_val <= alpha:
            clusters.append(index_cluster)
            cluster_pvals.append(p_val)

        if only_max_cluster:
            if p_sum > p_sum_max:
                clusters.clear()
                clusters.append(index_cluster)
                cluster_pvals = [p_val]
                p_sum_max = p_sum

    return cluster_pvals, clusters


def _null_distribution_2d(
    data_a: np.ndarray,
    data_b: np.ndarray | int | float,
    alpha: float,
    n_perm: int,
    two_tailed: bool,
    one_sample: bool,
    n_jobs: int = 1,
) -> np.ndarray:
    """Calculate null distribution of clusters.

    Parameters
    ----------
    data_a :  np.ndarray
        Data of three dimensions (first dimension is the number of
        measurements), e.g. shape: (n_subjects, n_freqs, n_times)
    alpha_ : float
        Significance level (p-value)
    n_perm_ : int
        No. of random permutations

    Returns
    -------
    null_distribution : np.ndarray
        Null distribution of shape (_n_perm, )
    """
    kwargs = {
        "data_a": data_a,
        "data_b": data_b,
        "alpha": alpha,
        "n_perm": n_perm,
        "two_tailed": two_tailed,
        "one_sample": one_sample,
    }
    if n_jobs in (0, 1):
        null_distr = [_single_p_sum_2d(**kwargs) for _ in range(n_perm)]
    else:
        from joblib import Parallel, delayed

        null_distr = Parallel(n_jobs=n_jobs, verbose=1)(
            delayed(_single_p_sum_2d)(**kwargs) for _ in range(n_perm)
        )
    return np.array(null_distr)


def _single_p_sum_2d(
    data_a: np.ndarray,
    data_b: np.ndarray | int | float,
    alpha: float,
    n_perm: int,
    two_tailed: bool,
    one_sample: bool,
) -> float:
    """"""
    if one_sample:
        sign = np.random.choice(
            a=np.array([-1, 1]), size=data_a.shape[0], replace=True
        ).reshape(data_a.shape[0], 1, 1)
        data_a_final = data_a.copy() * sign
    else:
        data_a_final = data_a

    p_values = pte_stats.permutation_2d(
        data_a=data_a_final,
        data_b=data_b,
        n_perm=n_perm,
        two_tailed=two_tailed,
    )
    p_values_inv = np.asarray(1 - p_values)
    labels_, num_clusters = measure.label(
        p_values <= alpha, return_num=True, connectivity=2
    )  # type: ignore

    p_sum_max = 0
    if num_clusters > 0:
        for i in range(num_clusters):
            index_cluster = np.asarray(labels_ == i + 1).nonzero()
            p_sum = np.sum(p_values_inv[index_cluster])
            p_sum_max = max(p_sum, p_sum_max)
    return p_sum_max


def clusters_from_pvals(
    p_vals: np.ndarray,
    alpha: float,
    correction_method: str,
    n_perm: int,
    min_cluster_size: int = 1,
) -> tuple[np.ndarray, int]:
    """Return significant clusters from array of p-values."""
    p_vals_signif = np.where(p_vals <= alpha)[0]
    if p_vals_signif.size == p_vals.size:
        return (np.ones_like(p_vals, dtype=np.int32), 1)
    if p_vals_signif.size > 1:
        p_vals_corr = pte_stats.correct_pvals(
            p_vals=p_vals,
            alpha=alpha,
            correction_method=correction_method,
            n_perm=n_perm,
        )
        if p_vals_corr.size > 1:
            clusters_raw = np.array(
                [1 if i in p_vals_corr else 0 for i, _ in enumerate(p_vals)]
            )
            clusters, cluster_count = get_clusters_1d(
                data=clusters_raw, min_cluster_size=min_cluster_size
            )
            return (clusters, cluster_count)
    return (np.ndarray([]), 0)


@njit
def get_clusters_1d(
    data: np.ndarray, min_cluster_size: int = 1
) -> tuple[np.ndarray, int]:
    """Cluster 1-D array of boolean values.

    Parameters
    ----------
    iterable : array-like of bool
        Array to be clustered.
    min_cluster_size : integer
        Minimum size of clusters to consider. Must be at least 1.

    Returns
    -------
    cluster_labels : np.array
        Array of shape (len(iterable), 1), where each value indicates the
        number of the cluster. Values are 0 if the item does not belong to
        a cluster
    cluster_count : int
        Number of detected cluster. Corresponds to the highest value in
        cluster_labels
    """
    min_cluster_size = max(min_cluster_size, 1)
    cluster_labels = np.zeros_like(data, dtype=np.int32)
    cluster_count = 0
    cluster_len = 0
    for idx, item in enumerate(data):
        if item:
            cluster_len += 1
            cluster_labels[idx] = cluster_count + 1
        else:
            if cluster_len >= min_cluster_size:
                cluster_count += 1
            else:
                cluster_labels[max(0, idx - cluster_len) : idx] = 0
            cluster_len = 0
    if cluster_len >= min_cluster_size:
        cluster_count += 1
    else:
        cluster_labels[min(-1, cluster_len) :] = 0
    return cluster_labels, cluster_count


@njit
def _null_distribution_from_pvals(
    p_values: np.ndarray,
    alpha: float = 0.05,
    n_perm: int = 1000,
    min_cluster_size: int = 1,
) -> np.ndarray:
    """Calculate null distribution of clusters from given p-values.

    Parameters
    ----------
    p_values :  np.ndarray
        Array of p-values
    alpha_ : float
        Significance level (p-value)
    n_perm_ : int
        No. of random permutations

    Returns
    -------
    null_distribution : np.ndarray
        Null distribution of shape (_n_perm, )
    """
    # loop through random permutation cycles
    null_distribution = np.zeros(n_perm)
    for i in range(n_perm):
        r_per = np.random.randint(
            low=0, high=p_values.shape[0], size=p_values.shape[0]
        )
        pvals_perm = p_values[r_per]
        labels_, n_clusters = get_clusters_1d(
            data=pvals_perm <= alpha, min_cluster_size=min_cluster_size
        )

        cluster_ind = {}
        if n_clusters == 0:
            null_distribution[i] = 0
        else:
            p_sum = np.zeros(n_clusters)
            for ind in range(n_clusters):
                cluster_ind[ind] = np.where(labels_ == ind + 1)[0]
                p_sum[ind] = np.sum(
                    np.asarray(1 - pvals_perm)[cluster_ind[ind]]
                )
            null_distribution[i] = np.max(p_sum)
    return null_distribution


@njit
def _null_distribution_1d_onesample(
    data_a: np.ndarray,
    data_b: int | float,
    alpha: float,
    n_perm: int,
    two_tailed: bool,
    min_cluster_size: int,
) -> np.ndarray:
    """Calculate null distribution of clusters.

    Parameters
    ----------
    data_a :  np.ndarray
        Data of three dimensions (first dimension is the number of
        measurements), e.g. shape: (n_subjects, n_freqs, n_times)
    alpha_ : float
        Significance level (p-value)
    n_perm_ : int
        No. of random permutations

    Returns
    -------
    null_distribution : np.ndarray
        Null distribution of shape (_n_perm, )
    """
    null_distr = np.empty((n_perm,))
    for n in range(n_perm):
        sign = np.random.choice(
            a=np.array([-1, 1]), size=data_a.shape[0], replace=True
        ).reshape(data_a.shape[0], 1)
        data_a_final = data_a.copy() * sign

        p_values = pte_stats.permutation_1d_onesample(
            data_a=data_a_final,
            data_b=data_b,
            n_perm=n_perm,
            two_tailed=two_tailed,
        )
        p_values_inv = np.asarray(1 - p_values)

        labels, num_clusters = get_clusters_1d(
            data=p_values <= alpha, min_cluster_size=min_cluster_size
        )

        p_sum_max = 0
        if num_clusters > 0:
            for i in range(num_clusters):
                index_cluster = np.asarray(labels == i + 1).nonzero()
                p_sum = np.sum(p_values_inv[index_cluster])
                p_sum_max = max(p_sum, p_sum_max)
        null_distr[n] = p_sum_max
    return null_distr


@njit
def _null_distribution_1d_twosample(
    data_a: np.ndarray,
    data_b: np.ndarray,
    alpha: float,
    n_perm: int,
    two_tailed: bool,
    min_cluster_size: int,
) -> np.ndarray:
    """Calculate null distribution of clusters.

    Parameters
    ----------
    data_a :  np.ndarray
        Data of three dimensions (first dimension is the number of
        measurements), e.g. shape: (n_subjects, n_freqs, n_times)
    alpha : float
        Significance level (p-value)
    n_perm  : int
        No. of random permutations

    Returns
    -------
    null_distribution : np.ndarray
        Null distribution of shape (_n_perm, )
    """
    null_distr = np.empty((n_perm,))
    for n in range(n_perm):
        p_values = pte_stats.permutation_1d_twosample(
            data_a=data_a,
            data_b=data_b,
            n_perm=n_perm,
            two_tailed=two_tailed,
        )
        p_values_inv = np.asarray(1 - p_values)

        labels, num_clusters = get_clusters_1d(
            data=p_values <= alpha, min_cluster_size=min_cluster_size
        )

        p_sum_max = 0
        if num_clusters > 0:
            for i in range(num_clusters):
                index_cluster = np.asarray(labels == i + 1).nonzero()
                p_sum = np.sum(p_values_inv[index_cluster])
                p_sum_max = max(p_sum, p_sum_max)
        null_distr[n] = p_sum_max
    return null_distr


def cluster_analysis_1d(
    data_a: np.ndarray,
    data_b: np.ndarray | int | float | None = None,
    alpha: float = 0.05,
    n_perm: int = 1000,
    only_max_cluster: bool = False,
    two_tailed: bool = True,
    min_cluster_size: int = 1,
) -> tuple[list, list]:
    """Calculate significant clusters and their corresponding p-values."""
    if data_b is None:
        data_b = 0
    if isinstance(data_b, (int, float)):
        f_perm = pte_stats.permutation_1d_onesample
        f_null_distr = _null_distribution_1d_onesample
    else:
        f_perm = pte_stats.permutation_1d_twosample
        f_null_distr = _null_distribution_1d_twosample

    p_values = f_perm(
        data_a=data_a,
        data_b=data_b,
        n_perm=n_perm,
        two_tailed=two_tailed,
    )

    labels, num_clusters = get_clusters_1d(
        data=p_values <= alpha, min_cluster_size=min_cluster_size
    )

    if num_clusters < 1:
        return [], []

    p_values_inv = np.asarray(1 - p_values)

    null_distr = f_null_distr(
        data_a=data_a,
        data_b=data_b,
        alpha=alpha,
        n_perm=n_perm,
        two_tailed=two_tailed,
        min_cluster_size=min_cluster_size,
    )

    clusters = []
    cluster_pvals = []
    p_sum_max = 0  # Is only used if only_max_cluster
    for i in range(num_clusters):
        # Cluster labels start at 1
        index_cluster = np.asarray(labels == i + 1).nonzero()
        p_sum = np.sum(p_values_inv[index_cluster])
        p_val = (n_perm - np.sum(p_sum >= null_distr) + 1) / (n_perm + 1)
        if p_val <= alpha:
            clusters.append(index_cluster)
            cluster_pvals.append(p_val)

        if only_max_cluster:
            if p_sum > p_sum_max:
                clusters.clear()
                clusters.append(index_cluster)
                cluster_pvals = [p_val]
                p_sum_max = p_sum

    return cluster_pvals, clusters


@njit
def cluster_analysis_from_pvals(
    p_values: np.ndarray,
    alpha: float = 0.05,
    n_perm: int = 1000,
    only_max_cluster: bool = False,
    min_cluster_size: int = 1,
) -> tuple[list, list]:
    """Calculate significant clusters and their corresponding p-values.

    Based on:
    https://github.com/neuromodulation/wjn_toolbox/blob/
    4745557040ad26f3b8498ca5d0c5d5dece2d3ba1/mypcluster.m

    https://garstats.wordpress.com/2018/09/06/cluster/

    Arguments
    ---------
    p_values :  numpy array
        Array of p-values. WARNING: MUST be one-dimensional
    alpha : float
        Significance level
    n_perm : int
        No. of random permutations for building cluster null-distribution
    only_max_cluster : bool, default = False
        Set to True to only return the most significant cluster.

    Returns
    -------
    cluster_pvals : list of float(s)
        List of p-values for each cluster
    clusters : list of numpy array(s)
        List of indices of each significant cluster
    """
    p_values_inv = np.asarray(1 - p_values)

    labels, num_clusters = get_clusters_1d(
        data=p_values <= alpha, min_cluster_size=min_cluster_size
    )

    null_distr = _null_distribution_from_pvals(p_values, alpha, n_perm)
    # Loop through clusters of p_val series or image
    clusters = []
    # Initialize empty list with specific data type for numba to work
    cluster_pvals = [np.float64(x) for x in range(0)]
    max_cluster_sum = 0
    # Cluster labels start at 1
    for cluster_i in range(num_clusters):
        index_cluster = np.where(labels == cluster_i + 1)[0]
        p_cluster_sum = np.sum(p_values_inv[index_cluster])
        p_val = (n_perm - np.sum(p_cluster_sum >= null_distr) + 1) / n_perm

        if p_val <= alpha:
            clusters.append(index_cluster)
            cluster_pvals.append(p_val)

        if only_max_cluster:
            if max_cluster_sum == 0 or p_cluster_sum > max_cluster_sum:
                clusters.clear()
                clusters.append(index_cluster)
                cluster_pvals = [p_val]
                max_cluster_sum = p_cluster_sum

    return cluster_pvals, clusters



"""Module for statistical analysis of time series."""
import numpy as np
from statsmodels.stats.multitest import fdrcorrection

import pte_stats


def timeseries_pvals(
    x: np.ndarray,
    y: int | float | np.ndarray,
    n_perm: int,
    two_tailed: bool,
):
    """Calculate sample-wise p-values for array using permutation testing."""
    p_vals = np.empty(len(x))
    if isinstance(y, (int, float)):
        for t_p, pred in enumerate(x):
            _, p_vals[t_p] = pte_stats.permutation_onesample(
                data_a=pred, data_b=y, n_perm=n_perm, two_tailed=two_tailed
            )
    else:
        for i, (x_, y_) in enumerate(zip(x, y)):
            _, p_vals[i] = pte_stats.permutation_twosample(
                data_a=x_, data_b=y_, n_perm=n_perm, two_tailed=two_tailed
            )
    return p_vals


def correct_pvals(
    p_vals: np.ndarray,
    alpha: float = 0.05,
    correction_method: str = "cluster",
    n_perm: int = 10000,
):
    """Correct p-values for multiple comparisons."""
    if correction_method == "cluster_pvals":
        _, signif = pte_stats.cluster_analysis_from_pvals(
            p_values=p_vals, alpha=alpha, n_perm=n_perm, only_max_cluster=False
        )
        if len(signif) > 0:
            signif = np.hstack(signif)
        else:
            signif = np.array([])
    elif correction_method == "fdr":
        rejected, _ = fdrcorrection(
            pvals=p_vals, alpha=alpha, method="poscorr", is_sorted=False
        )
        signif = np.where(rejected)[0]
    else:
        raise ValueError(
            "`correction_method` must be one of either `cluster_pvals` or"
            f"`fdr`. Got:{correction_method}."
        )
    return signif


def handle_baseline(
    baseline: None | tuple[int | float | None, int | float | None] = None,
    sfreq: int | float | None = None,
) -> tuple[int | None, int | None]:
    """Return baseline start and end indices."""
    if baseline is None:
        return None, None
    if any(baseline) and sfreq is None:
        raise ValueError(
            "If `baseline` is any value other than `None`, or `(None, None)`,"
            f" `sfreq` must be provided. Got: {baseline=}"
        )
    if not sfreq:
        sfreq = 0.0
    if baseline[0] is None:
        base_start = 0
    else:
        base_start = int(baseline[0] * sfreq)
    if baseline[1] is None:
        base_end = None
    else:
        base_end = int(baseline[1] * sfreq)
    return base_start, base_end


def handle_baseline_bytimes(
    baseline: None | tuple[int | float | None, int | float | None] = None,
    times: np.ndarray | None = None,
) -> tuple[int | None, int | None]:
    """Return baseline start and end indices."""
    if baseline is None:
        return None, None
    if any(baseline) and times is None:
        raise ValueError(
            "If `baseline` is any value other than `None`, or `(None, None)`,"
            f" `times` must be provided. Got: {baseline=}"
        )
    if baseline[0] is None:
        base_start = 0
    else:
        base_start = np.where(baseline[0] <= times)[0][0]
    if baseline[1] is None:
        base_end = None
    else:
        base_end = np.where(times <= baseline[1])[0][-1]
    return base_start, base_end


def baseline_correct(
    data: np.ndarray,
    baseline_mode: str = "percent",
    base_start: int | None = None,
    base_end: int | None = None,
    baseline_trialwise: bool = False,
) -> np.ndarray:
    """Baseline correct data."""
    if baseline_trialwise:
        axis = -1
    else:
        axis = (-2, -1)

    baseline = data[::, base_start:base_end]

    if baseline_mode == "percent":
        mean = np.mean(baseline, axis=axis, keepdims=True)
        data = (data - mean) / (mean) * 100
        return data
    if baseline_mode == "zscore":
        mean = np.mean(baseline, axis=axis, keepdims=True)
        data = (data - mean) / (np.std(baseline, axis=1, keepdims=True))
        return data
    if baseline_mode == "std":
        data /= np.std(baseline, axis=axis, keepdims=True)
        return data
    raise ValueError(
        "`baseline_mode` must be one of either `percent`, `std` or `zscore`."
        f" Got: {baseline_mode}."
    )



"""Module for permutation testing."""
import random

import numpy as np
from numba import njit
import pandas as pd
import scipy.stats


def spearmans_rho_permutation(
    x: np.ndarray |pd.Series, y: np.ndarray |pd.Series, n_perm: int = 5000
) -> tuple[float, float]:
    """Calculate permutation test for multiple repetitions of Spearmans Rho

    https://towardsdatascience.com/how-to-assess-statistical-significance-in-your-data-with-permutation-tests-8bb925b2113d

    Parameters
    ----------
    x (np array) : first distibution
    y (np array) : second distribution
    n_permp (int): number of permutations

    Returns
    -------
    gT (float) : estimated ground truth, here spearman's rho
    p_val (float) : p value of permutation test
    """

    # compute ground truth difference
    gT = scipy.stats.spearmanr(x, y)[0]
    #
    pV = np.array((x, y))
    # Initialize permutation:
    pD = []
    # Permutation loop:
    args_order = np.arange(0, pV.shape[1], 1)
    args_order_2 = np.arange(0, pV.shape[1], 1)
    for _ in range(n_perm):
        # Shuffle the data:
        random.shuffle(args_order)
        random.shuffle(args_order_2)
        # Compute permuted absolute difference of your two sampled
        # distributions and store it in pD:
        pD.append(
            scipy.stats.spearmanr(pV[0, args_order], pV[1, args_order_2])[0]
        )

    # calculate p value
    if gT < 0:
        p_val = (len(np.where(pD <= gT)[0]) + 1) / (n_perm + 1)
    else:
        p_val = (len(np.where(pD >= gT)[0]) + 1) / (n_perm + 1)

    return gT, p_val


def permutation_2d(
    data_a: np.ndarray,
    data_b: np.ndarray | int | float,
    n_perm: int = 1000,
    two_tailed: bool = True,
):
    """Perform permutation test with two-dimensional array.

    Parameters
    ----------
    x : array_like
        First distribution
    y : int or float
        Baseline against which to check for statistical significance
    n_perm : int
        Number of permutations
    two_tailed : bool, default: True
        Set to False if you would like to perform a one-sampled permutation
        test, else True

    Returns
    -------
    float
        Estimated difference of distribution from baseline
    float
        P-value of permutation test
    """
    if isinstance(data_b, (int, float)):
        data_b = np.ones((1, *data_a.shape[1:])) * data_b
        return permutation_2d_onesample(
            data_a=data_a, data_b=data_b, n_perm=n_perm, two_tailed=two_tailed
        )

    if data_a.shape[1:] != data_b.shape[1:]:
        raise ValueError(
            f"If `data_b` is an array, it must have the same shape as"
            f" `data_a`. Got: {data_a.shape=} and {data_b.shape=}."
        )
    return permutation_2d_twosample(
        data_a=data_a, data_b=data_b, n_perm=n_perm, two_tailed=two_tailed
    )


def permutation_1d(
    data_a: np.ndarray,
    data_b: np.ndarray | int | float,
    n_perm: int = 1000,
    two_tailed: bool = True,
) -> np.ndarray:
    """Perform permutation test with one-dimensional array.

    Parameters
    ----------
    data_a : array_like
        First distribution
    data_b : int or float
        Baseline against which to check for statistical significance
    n_perm : int
        Number of permutations
    two_tailed : bool, default: True
        Set to False if you would like to perform a one-sampled permutation
        test, else True

    Returns
    -------
    array
        P-values of permutation tests
    """
    if isinstance(data_b, (int, float)):
        return permutation_1d_onesample(
            data_a=data_a, data_b=data_b, n_perm=n_perm, two_tailed=two_tailed
        )
    return permutation_1d_twosample(
        data_a=data_a, data_b=data_b, n_perm=n_perm, two_tailed=two_tailed
    )


@njit
def permutation_1d_twosample(
    data_a: np.ndarray, data_b: np.ndarray, n_perm: int, two_tailed: bool
) -> np.ndarray:
    """"""
    p_values = np.empty((data_a.shape[1]))

    for i in np.arange(data_a.shape[1]):
        _, p = permutation_twosample(
            data_a[:, i], data_b[:, i], n_perm, two_tailed
        )
        p_values[i] = p
    return p_values


@njit
def permutation_1d_onesample(
    data_a: np.ndarray, data_b: int | float, n_perm: int, two_tailed: bool
) -> np.ndarray:
    """"""
    p_values = np.empty((data_a.shape[1]))

    for i in np.arange(data_a.shape[1]):
        _, p = permutation_onesample(data_a[:, i], data_b, n_perm, two_tailed)
        p_values[i] = p
    return p_values


@njit
def permutation_2d_twosample(
    data_a: np.ndarray, data_b: np.ndarray, n_perm: int, two_tailed: bool
) -> np.ndarray:
    """"""
    p_values = np.empty((data_a.shape[1], data_a.shape[2]))

    for i in np.arange(data_a.shape[1]):
        for j in np.arange(data_a.shape[2]):
            _, p = permutation_twosample(
                data_a[:, i, j], data_b[:, i, j], n_perm, two_tailed
            )
            p_values[i, j] = p
    return p_values


@njit
def permutation_2d_onesample(
    data_a: np.ndarray, data_b: np.ndarray, n_perm: int, two_tailed: bool
) -> np.ndarray:
    """"""
    p_values = np.empty((data_a.shape[1], data_a.shape[2]))

    for i in np.arange(data_a.shape[1]):
        for j in np.arange(data_a.shape[2]):
            _, p = permutation_onesample(
                data_a[:, i, j], data_b[:, i, j], n_perm, two_tailed
            )
            p_values[i, j] = p
    return p_values


@njit
def permutation_onesample(
    data_a: np.ndarray,
    data_b: int | float,
    n_perm: int = 10000,
    two_tailed: bool = True,
) -> tuple[float, float]:
    """Perform permutation test with one-sample distribution.

    Parameters
    ----------
    x : array_like
        First distribution
    y : int, float
        Baseline against which to check for statistical significance
    n_perm : int
        Number of permutations
    two_tailed : bool, default: True
        Set to False if you would like to perform a one-sampled permutation
        test, else True
    two_tailed : bool, default: True
        Set to False if you would like to perform a one-tailed permutation
        test, else True. If False, tests if data is significantly larger than
        baseline.

    Returns
    -------
    float
        Estimated difference of distribution from baseline
    float
        P-value of permutation test
    """

    zeroed = data_a - data_b
    p = np.empty(n_perm)
    z = np.mean(zeroed)
    if two_tailed:
        z = np.abs(z)
    # Run the simulation n_perm times
    for i in np.arange(n_perm):
        sign = np.random.choice(
            a=np.array([-1.0, 1.0]), size=len(data_a), replace=True
        )
        val_perm = np.mean(zeroed * sign)
        if two_tailed:
            val_perm = np.abs(val_perm)
        p[i] = val_perm

    # Compute effect size (Cohen's d)
    std = np.std(zeroed)
    abs_diff = np.abs(z)
    if std == 0.0:
        effect_size = 0.0
        if abs_diff != 0.0:
            print(
                "Warning: Pooled standard deviation of distributions was 0."
                " Cannot correctly calculate effect size (Cohen's d)."
            )
    else:
        effect_size = np.round(abs_diff / std, 3)

    return effect_size, (np.sum(p >= z) + 1) / (n_perm + 1)


@njit
def permutation_twosample(
    data_a: np.ndarray,
    data_b: np.ndarray,
    n_perm: int = 10000,
    two_tailed: bool = True,
) -> tuple[float, float]:
    """Perform permutation test.

    Parameters
    ----------
    x : array_like
        First distribution
    y : array_like
        Second distribution
    n_perm : int
        Number of permutations
    two_tailed : bool, default: True
        Set to False if you would like to perform a one-sampled permutation
        test, else True
    two_tailed : bool, default: True
        Set to False if you would like to perform a one-tailed permutation
        test, else True

    Returns
    -------
    float
        Estimated difference of distribution means
    float
        P-value of permutation test
    """
    if two_tailed:
        zeroed = np.abs(np.mean(data_a) - np.mean(data_b))
        data = np.concatenate((data_a, data_b), axis=0)
        half = int(len(data) / 2)
        p = np.empty(n_perm)
        for i in np.arange(0, n_perm):
            np.random.shuffle(data)
            # Compute permuted absolute difference of the two sampled
            # distributions
            p[i] = np.abs(np.mean(data[:half]) - np.mean(data[half:]))
    else:
        zeroed = np.mean(data_a) - np.mean(data_b)
        data = np.concatenate((data_a, data_b), axis=0)
        half = int(len(data) / 2)
        p = np.empty(n_perm)
        for i in np.arange(0, n_perm):
            np.random.shuffle(data)
            # Compute permuted absolute difference of the two sampled
            # distributions
            p[i] = np.mean(data[:half]) - np.mean(data[half:])

    # Compute effect size (Cohen's d)
    n_a = data_a.size
    n_b = data_b.size
    pooled_std = np.sqrt(
        (
            ((n_a - 1) * np.square(np.std(data_a)))
            + ((n_b - 1) * np.square(np.std(data_b)))
        )
        / (n_a + n_b - 2)
    )
    effect_size = np.abs(np.mean(data_a) - np.mean(data_b)) / pooled_std
    effect_size = np.round(effect_size, 3)
    return effect_size, (np.sum(p >= zeroed) + 1) / (n_perm + 1)
